@@ -1,53 +1,62 @@
 from __future__ import annotations
 
-import os
-import sys
 from codecs import decode
-from collections.abc import Mapping
-from contextlib import AbstractContextManager
-from dataclasses import dataclass, field
-from functools import singledispatch, wraps
+from collections.abc import Mapping, Sequence
+from enum import Enum, auto
+from functools import cache, partial, singledispatch
+from importlib.metadata import version
 from importlib.util import find_spec
-from inspect import Parameter, signature
-from pathlib import Path
 from types import EllipsisType
-from typing import TYPE_CHECKING, TypeVar
+from typing import TYPE_CHECKING, TypeVar, overload
 from warnings import warn
 
 import h5py
 import numpy as np
 import pandas as pd
 import scipy
-import scipy.sparse
+from numpy.typing import NDArray
 from packaging.version import Version
-
-from .exceptiongroups import add_note  # noqa: F401
+from zarr import Array as ZarrArray  # noqa: F401
+from zarr import Group as ZarrGroup
 
 if TYPE_CHECKING:
     from typing import Any
 
+
 #############################
 # scipy sparse array comapt #
 #############################
 
 
-CAN_USE_SPARSE_ARRAY = Version(scipy.__version__) >= Version("1.11")
-
-if not CAN_USE_SPARSE_ARRAY:
-
-    class SpArray:
-        @staticmethod
-        def __repr__():
-            return "mock scipy.sparse.sparray"
-else:
-    SpArray = scipy.sparse.sparray
+CSMatrix = scipy.sparse.csr_matrix | scipy.sparse.csc_matrix
+CSArray = scipy.sparse.csr_array | scipy.sparse.csc_array
 
 
-class Empty:
-    pass
+class Empty(Enum):
+    TOKEN = auto()
 
 
-Index1D = slice | int | str | np.int64 | np.ndarray
+Index1DNorm = slice | NDArray[np.bool_] | NDArray[np.integer]
+# TODO: pd.Index[???]
+Index1D = (
+    # 0D index
+    int
+    | str
+    | np.int64
+    # normalized 1D idex
+    | Index1DNorm
+    # different containers for mask, obs/varnames, or numerical index
+    | Sequence[int]
+    | Sequence[str]
+    | Sequence[bool]
+    | pd.Series  # bool, int, str
+    | pd.Index
+    | pd.api.extensions.ExtensionArray  # bool | int | str
+    | NDArray[np.str_]
+    | np.matrix  # bool
+    | CSMatrix  # bool
+    | CSArray  # bool
+)
 IndexRest = Index1D | EllipsisType
 Index = (
     IndexRest
@@ -56,54 +65,45 @@ Index = (
     | tuple[Index1D, Index1D, EllipsisType]
     | tuple[EllipsisType, Index1D, Index1D]
     | tuple[Index1D, EllipsisType, Index1D]
-    | scipy.sparse.spmatrix
-    | SpArray
+    | CSMatrix
+    | CSArray
 )
 H5Group = h5py.Group
 H5Array = h5py.Dataset
 H5File = h5py.File
 
-
-#############################
-# stdlib
-#############################
-
-
-if sys.version_info >= (3, 11):
-    from contextlib import chdir
+# h5py recommends using .astype("T") over .asstr() when using numpy ≥2
+if TYPE_CHECKING:
+    from h5py._hl.dataset import AsStrView as H5AsStrView
+    from h5py._hl.dataset import AsTypeView as H5AsTypeView
 else:
-
-    @dataclass
-    class chdir(AbstractContextManager):
-        path: Path
-        _old_cwd: list[Path] = field(default_factory=list)
-
-        def __enter__(self) -> None:
-            self._old_cwd.append(Path())
-            os.chdir(self.path)
-
-        def __exit__(self, *_exc_info) -> None:
-            os.chdir(self._old_cwd.pop())
+    try:
+        try:
+            from h5py._hl.dataset import AsStrView as H5AsStrView
+            from h5py._hl.dataset import AsTypeView as H5AsTypeView
+        except ImportError:
+            # h5py 3.11 uses AsStrWrapper/AstypeWrapper (lowercase 't')
+            from h5py._hl.dataset import AsStrWrapper as H5AsStrView
+            from h5py._hl.dataset import AstypeWrapper as H5AsTypeView
+    except ImportError:  # pragma: no cover
+        warn("AsTypeView changed import location", DeprecationWarning, stacklevel=1)
+        _ds = h5py.File.in_memory().create_dataset("x", shape=(), dtype="S1")
+        H5AsStrView = type(_ds.asstr())
+        H5AsTypeView = type(_ds.astype("U1"))
+        del _ds
 
 
 #############################
 # Optional deps
 #############################
+@cache
+def is_zarr_v2() -> bool:
+    return Version(version("zarr")) < Version("3.0.0")
 
-if find_spec("zarr") or TYPE_CHECKING:
-    from zarr.core import Array as ZarrArray
-    from zarr.hierarchy import Group as ZarrGroup
-else:
 
-    class ZarrArray:
-        @staticmethod
-        def __repr__():
-            return "mock zarr.core.Array"
-
-    class ZarrGroup:
-        @staticmethod
-        def __repr__():
-            return "mock zarr.core.Group"
+if is_zarr_v2():
+    msg = "anndata will no longer support zarr v2 in the near future. Please prepare to upgrade to zarr>=3."
+    warn(msg, DeprecationWarning, stacklevel=2)
 
 
 if find_spec("awkward") or TYPE_CHECKING:
@@ -140,7 +140,47 @@ else:
             return "mock dask.array.core.Array"
 
 
-if find_spec("cupy") or TYPE_CHECKING:
+if find_spec("xarray") or TYPE_CHECKING:
+    import xarray
+    from xarray import DataArray as XDataArray
+    from xarray import Dataset as XDataset
+    from xarray import Variable as XVariable
+    from xarray.backends import BackendArray as XBackendArray
+    from xarray.backends.zarr import ZarrArrayWrapper as XZarrArrayWrapper
+else:
+    xarray = None
+
+    class XDataArray:
+        def __repr__(self) -> str:
+            return "mock DataArray"
+
+    class XDataset:
+        def __repr__(self) -> str:
+            return "mock Dataset"
+
+    class XVariable:
+        def __repr__(self) -> str:
+            return "mock Variable"
+
+    class XZarrArrayWrapper:
+        def __repr__(self) -> str:
+            return "mock ZarrArrayWrapper"
+
+    class XBackendArray:
+        def __repr__(self) -> str:
+            return "mock BackendArray"
+
+
+# https://github.com/scverse/anndata/issues/1749
+def is_cupy_importable() -> bool:
+    try:
+        import cupy  # noqa: F401
+    except ImportError:
+        return False
+    return True
+
+
+if is_cupy_importable() or TYPE_CHECKING:
     from cupy import ndarray as CupyArray
     from cupyx.scipy.sparse import csc_matrix as CupyCSCMatrix
     from cupyx.scipy.sparse import csr_matrix as CupyCSRMatrix
@@ -176,14 +216,99 @@ else:
             return "mock cupy.ndarray"
 
 
+if find_spec("legacy_api_wrap") or TYPE_CHECKING:
+    from legacy_api_wrap import legacy_api  # noqa: TID251
+
+    old_positionals = partial(legacy_api, category=FutureWarning)
+else:
+
+    def old_positionals(*old_positionals):
+        return lambda func: func
+
+
 #############################
 # IO helpers
 #############################
 
+NUMPY_2 = Version(version("numpy")) >= Version("2")
+
+NULLABLE_NUMPY_STRING_TYPE = (
+    np.dtypes.StringDType(na_object=pd.NA) if NUMPY_2 else np.dtype("O")
+)
+
+PANDAS_SUPPORTS_NA_VALUE = Version(version("pandas")) >= Version("2.3")
+
+
+PANDAS_STRING_ARRAY_TYPES: list[type[pd.api.extensions.ExtensionArray]] = [
+    pd.arrays.StringArray,
+    pd.arrays.ArrowStringArray,
+]
+# these are removed in favor of the above classes: https://github.com/pandas-dev/pandas/pull/62149
+try:
+    from pandas.core.arrays.string_ import StringArrayNumpySemantics
+except ImportError:
+    pass
+else:
+    PANDAS_STRING_ARRAY_TYPES += [StringArrayNumpySemantics]
+try:
+    from pandas.core.arrays.string_arrow import ArrowStringArrayNumpySemantics
+except ImportError:
+    pass
+else:
+    PANDAS_STRING_ARRAY_TYPES += [ArrowStringArrayNumpySemantics]
+
+
+@overload
+def pandas_as_str(a: pd.Index[Any]) -> pd.Index[str]: ...
+@overload
+def pandas_as_str(a: pd.Series[Any]) -> pd.Series[str]: ...
+
+
+def pandas_as_str(a: pd.Index | pd.Series) -> pd.Index[str] | pd.Series[str]:
+    """Convert to fitting dtype, maintaining NA semantics if possible.
+
+    This is `"str"` when `pd.options.future.infer_string` is `True` (e.g. in Pandas 3+), and `"object"` otherwise.
+    """
+    if not pd.options.future.infer_string:
+        return a.astype(str)
+    if a.array.dtype == "string":  # any `pd.StringDtype`
+        return a
+    if PANDAS_SUPPORTS_NA_VALUE:
+        dtype = pd.StringDtype(na_value=a.array.dtype.na_value)
+    elif a.array.dtype.na_value is pd.NA:
+        dtype = pd.StringDtype()  # NA semantics
+    elif a.array.dtype.na_value is np.nan and find_spec("pyarrow"):  # noqa: PLW0177
+        # on pandas 2.2, this is the only way to get `np.nan` semantics
+        dtype = pd.StringDtype("pyarrow_numpy")
+    else:
+        msg = (
+            f"Converting an array with `dtype.na_value={a.array.dtype.na_value}` to a string array requires pyarrow or pandas>=2.3. "
+            "Converting to `pd.NA` semantics instead."
+        )
+        warn(msg, UserWarning, stacklevel=2)
+        dtype = pd.StringDtype()  # NA semantics
+    return a.astype(dtype)
+
+
+V = TypeVar("V")
+T = TypeVar("T")
+
+
+@overload
+def _read_attr(
+    attrs: Mapping[str, V], name: str, default: Empty = Empty.TOKEN
+) -> V: ...
+
+
+@overload
+def _read_attr(attrs: Mapping[str, V], name: str, default: T) -> V | T: ...
+
 
 @singledispatch
-def _read_attr(attrs: Mapping, name: str, default: Any | None = Empty):
-    if default is Empty:
+def _read_attr(
+    attrs: Mapping[str, V], name: str, default: T | Empty = Empty.TOKEN
+) -> V | T:
+    if default is Empty.TOKEN:
         return attrs[name]
     else:
         return attrs.get(name, default=default)
@@ -191,8 +316,8 @@ def _read_attr(attrs: Mapping, name: str, default: Any | None = Empty):
 
 @_read_attr.register(h5py.AttributeManager)
 def _read_attr_hdf5(
-    attrs: h5py.AttributeManager, name: str, default: Any | None = Empty
-):
+    attrs: h5py.AttributeManager, name: str, default: T | Empty = Empty.TOKEN
+) -> str | T:
     """
     Read an HDF5 attribute and perform all necessary conversions.
 
@@ -201,20 +326,17 @@ def _read_attr_hdf5(
     For example Julia's HDF5.jl writes string attributes as fixed-size strings, which
     are read as bytes by h5py.
     """
-    if name not in attrs and default is not Empty:
+    if name not in attrs and default is not Empty.TOKEN:
         return default
     attr = attrs[name]
     attr_id = attrs.get_id(name)
     dtype = h5py.check_string_dtype(attr_id.dtype)
-    if dtype is None:
+    if dtype is None or dtype.length is None:  # variable-length string, no problem
         return attr
-    else:
-        if dtype.length is None:  # variable-length string, no problem
-            return attr
-        elif len(attr_id.shape) == 0:  # Python bytestring
-            return attr.decode("utf-8")
-        else:  # NumPy array
-            return [decode(s, "utf-8") for s in attr]
+    elif len(attr_id.shape) == 0:  # Python bytestring
+        return attr.decode("utf-8")
+    else:  # NumPy array
+        return [decode(s, "utf-8") for s in attr]
 
 
 def _from_fixed_length_strings(value):
@@ -242,7 +364,7 @@ def _from_fixed_length_strings(value):
 
 
 def _decode_structured_array(
-    arr: np.ndarray, dtype: np.dtype | None = None, copy: bool = False
+    arr: np.ndarray, *, dtype: np.dtype | None = None, copy: bool = False
 ) -> np.ndarray:
     """
     h5py 3.0 now reads all strings as bytes. There is a helper method which can convert these to strings,
@@ -273,8 +395,12 @@ def _to_fixed_length_strings(value: np.ndarray) -> np.ndarray:
     """\
     Convert variable length strings to fixed length.
 
-    Currently a workaround for
-    https://github.com/zarr-developers/zarr-python/pull/422
+    Formerly a workaround for
+    https://github.com/zarr-developers/zarr-python/pull/422,
+    resolved in https://github.com/zarr-developers/zarr-python/pull/813.
+
+    But if we didn't do this conversion, we would have to use a special codec in v2
+    for objects and v3 doesn't support objects at all.  So we leave this function as-is.
     """
     new_dtype = []
     for dt_name, (dt_type, dt_offset) in value.dtype.fields.items():
@@ -297,7 +423,7 @@ def _require_group_write_dataframe(
 ) -> Group_T:
     if len(df.columns) > 5_000 and isinstance(f, H5Group):
         # actually 64kb is the limit, but this should be a conservative estimate
-        return f.create_group(name, track_order=True, *args, **kwargs)
+        return f.create_group(name, *args, track_order=True, **kwargs)
     return f.require_group(name, *args, **kwargs)
 
 
@@ -346,11 +472,12 @@ def _move_adj_mtx(d):
             and isinstance(n[k], scipy.sparse.spmatrix | np.ndarray)
             and len(n[k].shape) == 2
         ):
-            warn(
-                f"Moving element from .uns['neighbors']['{k}'] to .obsp['{k}'].\n\n"
-                "This is where adjacency matrices should go now.",
-                FutureWarning,
+            msg = (
+                f"Moving element from .uns['neighbors'][{k!r}] to .obsp[{k!r}].\n\n"
+                "This is where adjacency matrices should go now."
             )
+            # 5: caller -> 4: legacy_api_wrap -> 3: `AnnData.__init__` -> 2: `_init_as_actual` → 1: here
+            warn(msg, FutureWarning, stacklevel=5)
             obsp[k] = n.pop(k)
 
 
@@ -364,60 +491,6 @@ def _find_sparse_matrices(d: Mapping, n: int, keys: tuple, paths: list):
     return paths
 
 
-# This function was adapted from scikit-learn
-# github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/validation.py
-def _deprecate_positional_args(func=None, *, version: str = "1.0 (renaming of 0.25)"):
-    """Decorator for methods that issues warnings for positional arguments.
-    Using the keyword-only argument syntax in pep 3102, arguments after the
-    * will issue a warning when passed as a positional argument.
-
-    Parameters
-    ----------
-    func
-        Function to check arguments on.
-    version
-        The version when positional arguments will result in error.
-    """
-
-    def _inner_deprecate_positional_args(f):
-        sig = signature(f)
-        kwonly_args = []
-        all_args = []
-
-        for name, param in sig.parameters.items():
-            if param.kind == Parameter.POSITIONAL_OR_KEYWORD:
-                all_args.append(name)
-            elif param.kind == Parameter.KEYWORD_ONLY:
-                kwonly_args.append(name)
-
-        @wraps(f)
-        def inner_f(*args, **kwargs):
-            extra_args = len(args) - len(all_args)
-            if extra_args <= 0:
-                return f(*args, **kwargs)
-
-            # extra_args > 0
-            args_msg = [
-                f"{name}={arg}"
-                for name, arg in zip(kwonly_args[:extra_args], args[-extra_args:])
-            ]
-            args_msg = ", ".join(args_msg)
-            warn(
-                f"Pass {args_msg} as keyword args. From version {version} passing "
-                "these as positional arguments will result in an error",
-                FutureWarning,
-            )
-            kwargs.update(zip(sig.parameters, args))
-            return f(**kwargs)
-
-        return inner_f
-
-    if func is not None:
-        return _inner_deprecate_positional_args(func)
-
-    return _inner_deprecate_positional_args
-
-
 def _transpose_by_block(dask_array: DaskArray) -> DaskArray:
     import dask.array as da
 
@@ -443,11 +516,3 @@ def _safe_transpose(x):
         return _transpose_by_block(x)
     else:
         return x.T
-
-
-def _map_cat_to_str(cat: pd.Categorical) -> pd.Categorical:
-    if Version(pd.__version__) >= Version("2.1"):
-        # Argument added in pandas 2.1
-        return cat.map(str, na_action="ignore")
-    else:
-        return cat.map(str)
