@@ -1,7 +1,9 @@
 from __future__ import annotations
 
-from contextlib import ExitStack
+from contextlib import nullcontext
 from copy import deepcopy
+from functools import partial
+from importlib.metadata import version
 from operator import mul
 from typing import TYPE_CHECKING
 
@@ -22,7 +24,7 @@ from anndata._core.views import (
     SparseCSRArrayView,
     SparseCSRMatrixView,
 )
-from anndata.compat import CAN_USE_SPARSE_ARRAY, CupyCSCMatrix, DaskArray
+from anndata.compat import NUMPY_2, CSArray, CupyCSCMatrix, DaskArray
 from anndata.tests.helpers import (
     BASE_MATRIX_PARAMS,
     CUPY_MATRIX_PARAMS,
@@ -30,14 +32,19 @@ from anndata.tests.helpers import (
     GEN_ADATA_DASK_ARGS,
     assert_equal,
     gen_adata,
+    single_int_subset,
     single_subset,
-    slice_subset,
+    slice_int_subset,
     subset_func,
 )
 from anndata.utils import asarray
 
 if TYPE_CHECKING:
-    from types import EllipsisType
+    from collections.abc import Callable
+    from types import EllipsisType, FunctionType
+
+    from anndata.compat import Index
+
 
 IGNORE_SPARSE_EFFICIENCY_WARNING = pytest.mark.filterwarnings(
     "ignore:Changing the sparsity structure:scipy.sparse.SparseEfficiencyWarning"
@@ -70,31 +77,44 @@ class NDArraySubclass(np.ndarray):
 
 
 @pytest.fixture
-def adata():
+def adata() -> ad.AnnData:
     adata = ad.AnnData(np.zeros((100, 100)))
     adata.obsm["o"] = np.zeros((100, 50))
     adata.varm["o"] = np.zeros((100, 50))
     return adata
 
 
+@pytest.fixture(scope="session")
+def adata_gen_session(matrix_type) -> ad.AnnData:
+    adata = gen_adata((30, 15), X_type=matrix_type)
+    adata.raw = adata.copy()
+    return adata
+
+
+@pytest.fixture
+def adata_gen(adata_gen_session: ad.AnnData) -> ad.AnnData:
+    return adata_gen_session.copy()
+
+
 @pytest.fixture(
     params=BASE_MATRIX_PARAMS + DASK_MATRIX_PARAMS + CUPY_MATRIX_PARAMS,
+    scope="session",
 )
 def matrix_type(request):
     return request.param
 
 
-@pytest.fixture(params=BASE_MATRIX_PARAMS + DASK_MATRIX_PARAMS)
+@pytest.fixture(params=BASE_MATRIX_PARAMS + DASK_MATRIX_PARAMS, scope="session")
 def matrix_type_no_gpu(request):
     return request.param
 
 
-@pytest.fixture(params=BASE_MATRIX_PARAMS)
+@pytest.fixture(params=BASE_MATRIX_PARAMS, scope="session")
 def matrix_type_base(request):
     return request.param
 
 
-@pytest.fixture(params=["layers", "obsm", "varm"])
+@pytest.fixture(params=["layers", "obsm", "varm"], scope="session")
 def mapping_name(request):
     return request.param
 
@@ -130,20 +150,7 @@ def test_convert_error():
     adata = ad.AnnData(np.array([[1, 2], [3, 0]]))
     no_array = [[1], []]
 
-    if Version(np.__version__) >= Version("1.24"):
-        stack = pytest.raises(ValueError, match=r"Failed to convert")
-    else:
-        stack = ExitStack()
-        stack.enter_context(
-            pytest.warns(
-                np.VisibleDeprecationWarning,
-                match=r"ndarray from ragged.*is deprecated",
-            )
-        )
-        stack.enter_context(
-            pytest.raises(ValueError, match=r"setting an array element with a sequence")
-        )
-    with stack:
+    with pytest.raises(ValueError, match=r"Failed to convert"):
         adata[:, 0].X = no_array
 
 
@@ -151,8 +158,8 @@ def test_view_subset_shapes():
     adata = gen_adata((20, 10), **GEN_ADATA_DASK_ARGS)
 
     view = adata[:, ::2]
-    assert view.var.shape == (5, 8)
-    assert {k: v.shape[0] for k, v in view.varm.items()} == {k: 5 for k in view.varm}
+    assert view.var.shape == (5, adata.var.shape[1])
+    assert {k: v.shape[0] for k, v in view.varm.items()} == dict.fromkeys(view.varm, 5)
 
 
 def test_modify_view_component(matrix_type, mapping_name, request):
@@ -175,14 +182,27 @@ def test_modify_view_component(matrix_type, mapping_name, request):
     with pytest.warns(ad.ImplicitModificationWarning, match=rf".*\.{mapping_name}.*"):
         m[0, 0] = 100
     assert not subset.is_view
-    assert getattr(subset, mapping_name)["m"][0, 0] == 100
+    # TODO: Remove `raises` after https://github.com/scipy/scipy/pull/23626 becomes minimum version i.e., scipy 1.17.
+
+    is_dask_with_broken_view_setting = (
+        "sparse_dask" in request.node.callspec.id
+        and Version(version("dask")) >= Version("2025.02.0")
+    )
+    is_sparse_array_in_lower_dask_version = (
+        not is_dask_with_broken_view_setting
+        and isinstance(m, DaskArray)
+        and isinstance(m._meta, CSArray)
+    )
+    with (
+        pytest.raises(ValueError, match=r"shape mismatch")
+        if Version(version("scipy")) < Version("1.17.0rc0")
+        and (is_sparse_array_in_lower_dask_version or is_dask_with_broken_view_setting)
+        else nullcontext()
+    ):
+        assert getattr(subset, mapping_name)["m"][0, 0] == 100
 
     assert init_hash == hash_func(adata)
 
-    if "sparse_array_dask_array" in request.node.callspec.id and CAN_USE_SPARSE_ARRAY:
-        msg = "sparse arrays in dask are generally expected to fail but in this case they do not"
-        pytest.fail(msg)
-
 
 @pytest.mark.parametrize("attr", ["obsm", "varm"])
 def test_set_obsm_key(adata, attr):
@@ -316,7 +336,7 @@ def test_not_set_subset_X(matrix_type_base, subset_func):
     init_hash = joblib.hash(adata)
     orig_X_val = adata.X.copy()
     while True:
-        subset_idx = slice_subset(adata.obs_names)
+        subset_idx = slice_int_subset(adata.obs_names)
         if len(adata[subset_idx, :]) > 2:
             break
     subset = adata[subset_idx, :]
@@ -330,7 +350,7 @@ def test_not_set_subset_X(matrix_type_base, subset_func):
     with pytest.warns(ad.ImplicitModificationWarning, match=r".*X.*"):
         subset.X[:, internal_idx] = 1
     assert not subset.is_view
-    assert not np.any(asarray(adata.X != orig_X_val))
+    assert not np.any(asarray(orig_X_val != adata.X))
 
     assert init_hash == joblib.hash(adata)
     assert isinstance(subset.X, type(adata.X))
@@ -344,7 +364,7 @@ def test_not_set_subset_X_dask(matrix_type_no_gpu, subset_func):
     init_hash = tokenize(adata)
     orig_X_val = adata.X.copy()
     while True:
-        subset_idx = slice_subset(adata.obs_names)
+        subset_idx = slice_int_subset(adata.obs_names)
         if len(adata[subset_idx, :]) > 2:
             break
     subset = adata[subset_idx, :]
@@ -358,7 +378,7 @@ def test_not_set_subset_X_dask(matrix_type_no_gpu, subset_func):
     with pytest.warns(ad.ImplicitModificationWarning, match=r".*X.*"):
         subset.X[:, internal_idx] = 1
     assert not subset.is_view
-    assert not np.any(asarray(adata.X != orig_X_val))
+    assert not np.any(asarray(orig_X_val != adata.X))
 
     assert init_hash == tokenize(adata)
     assert isinstance(subset.X, type(adata.X))
@@ -392,7 +412,7 @@ def test_set_subset_obsm(adata, subset_func):
     orig_obsm_val = adata.obsm["o"].copy()
 
     while True:
-        subset_idx = slice_subset(adata.obs_names)
+        subset_idx = slice_int_subset(adata.obs_names)
         if len(adata[subset_idx, :]) > 2:
             break
     subset = adata[subset_idx, :]
@@ -415,7 +435,7 @@ def test_set_subset_varm(adata, subset_func):
     orig_varm_val = adata.varm["o"].copy()
 
     while True:
-        subset_idx = slice_subset(adata.var_names)
+        subset_idx = slice_int_subset(adata.var_names)
         if (adata[:, subset_idx]).shape[1] > 2:
             break
     subset = adata[:, subset_idx]
@@ -526,27 +546,31 @@ def test_layers_view():
     assert view_hash != joblib.hash(view_adata)
 
 
-# TODO: This can be flaky. Make that stop
-def test_view_of_view(matrix_type, subset_func, subset_func2):
-    adata = gen_adata((30, 15), X_type=matrix_type)
-    adata.raw = adata.copy()
-    if subset_func is single_subset:
+# TODO: less combinatoric; split up into 2 tests:
+# 1. each subset func produces the right `oidx`/`vidx` kind (slice, array[int], array[bool])
+# 2. each `oidx`/`vidx` kind can be sliced with each subset func
+# going from #subset_func² to #subset_func × 3 {ov}idx kinds × 2 tests
+def test_view_of_view(adata_gen: ad.AnnData, subset_func, subset_func2) -> None:
+    adata = adata_gen
+    if subset_func in {single_subset, single_int_subset}:
         pytest.xfail("Other subset generating functions have trouble with this")
     var_s1 = subset_func(adata.var_names, min_size=4)
     var_view1 = adata[:, var_s1]
-    adata[:, var_s1].X
+    adata[:, var_s1].X  # noqa: B018
     var_s2 = subset_func2(var_view1.var_names)
     var_view2 = var_view1[:, var_s2]
+
     assert var_view2._adata_ref is adata
     assert isinstance(var_view2.X, type(adata.X))
+
     obs_s1 = subset_func(adata.obs_names, min_size=4)
     obs_view1 = adata[obs_s1, :]
     obs_s2 = subset_func2(obs_view1.obs_names)
+
     assert adata[obs_s1, :][:, var_s1][obs_s2, :]._adata_ref is adata
     assert isinstance(obs_view1.X, type(adata.X))
 
     view_of_actual_copy = adata[:, var_s1].copy()[obs_s1, :].copy()[:, var_s2].copy()
-
     view_of_view_copy = adata[:, var_s1][obs_s1, :][:, var_s2].copy()
 
     assert_equal(view_of_actual_copy, view_of_view_copy, exact=True)
@@ -631,7 +655,7 @@ def test_invalid_scalar_index(adata, index):
 
 @pytest.mark.parametrize("obs", [False, True])
 @pytest.mark.parametrize("index", [-100, -50, -1])
-def test_negative_scalar_index(adata, index: int, obs: bool):
+def test_negative_scalar_index(*, adata, index: int, obs: bool):
     pos_index = index + (adata.n_obs if obs else adata.n_vars)
 
     if obs:
@@ -673,9 +697,7 @@ def test_viewness_propagation_allclose(adata):
     assert np.allclose(a.varm["o"], b.varm["o"].copy(), equal_nan=True)
 
 
-spmat = [sparse.csr_matrix, sparse.csc_matrix]
-if CAN_USE_SPARSE_ARRAY:
-    spmat += [sparse.csr_array, sparse.csc_array]
+spmat = [sparse.csr_matrix, sparse.csc_matrix, sparse.csr_array, sparse.csc_array]
 
 
 @pytest.mark.parametrize("spmat", spmat)
@@ -693,10 +715,7 @@ def test_deepcopy_subset(adata, spmat: type):
     view_type = (
         SparseCSRMatrixView if spmat is sparse.csr_matrix else SparseCSCMatrixView
     )
-    if CAN_USE_SPARSE_ARRAY:
-        view_type = (
-            SparseCSRArrayView if spmat is sparse.csr_array else SparseCSCArrayView
-        )
+    view_type = SparseCSRArrayView if spmat is sparse.csr_array else SparseCSCArrayView
     assert not isinstance(
         adata.obsp["spmat"],
         view_type,
@@ -704,9 +723,13 @@ def test_deepcopy_subset(adata, spmat: type):
     np.testing.assert_array_equal(adata.obsp["spmat"].shape, (10, 10))
 
 
-array_type = [asarray, sparse.csr_matrix, sparse.csc_matrix]
-if CAN_USE_SPARSE_ARRAY:
-    array_type += [sparse.csr_array, sparse.csc_array]
+array_type = [
+    asarray,
+    sparse.csr_matrix,
+    sparse.csc_matrix,
+    sparse.csr_array,
+    sparse.csc_array,
+]
 
 
 # https://github.com/scverse/anndata/issues/680
@@ -726,12 +749,7 @@ def test_view_mixin_copies_data(adata, array_type: type, attr):
         getattr(adata, attr)["arr"] = X
 
     view = adata[:50]
-
-    if attr == "X":
-        arr_view = view.X
-    else:
-        arr_view = getattr(view, attr)["arr"]
-
+    arr_view = view.X if attr == "X" else getattr(view, attr)["arr"]
     arr_view_copy = arr_view.copy()
 
     if sparse.issparse(X):
@@ -790,6 +808,56 @@ def test_dataframe_view_index_setting():
     assert a2.obs.index.values.tolist() == ["a", "b"]
 
 
+def _n(t: type | FunctionType) -> str:
+    match t.__module__.split(".")[0], t.__name__:
+        case "pandas", name:
+            return f"pd.{name}"
+        case "numpy", name:
+            return f"np.{name}"
+        case mod, name:
+            return f"{mod}.{name}"
+
+
+def _string_dtypes(t: type | FunctionType) -> list[str]:
+    if t.__module__.startswith("pandas"):
+        return ["string[python]", "string[pyarrow]"]
+    if NUMPY_2:
+        return ["T"]
+    return []
+
+
+# only test 1D since 2D is tested by all the other tests
+@pytest.mark.parametrize(
+    ("mk_idx", "expected"),
+    [
+        pytest.param(lambda: slice(None), [[1, 2], [3, 4], [5, 6]], id="slice-all"),
+        pytest.param(lambda: slice(2), [[1, 2], [3, 4]], id="slice-some"),
+        pytest.param(lambda: [0], [[1, 2]], id="list-int"),
+        pytest.param(lambda: ["0"], [[1, 2]], id="list-str"),
+        pytest.param(lambda: [False, False, True], [[5, 6]], id="list-bool"),
+        *(
+            pytest.param(
+                partial(t, [0, 2], dtype=dt), [[1, 2], [5, 6]], id=f"{_n(t)}-{dt}"
+            )
+            for t in [np.array, pd.Index, pd.Series, pd.array]
+            for dt in ["int", *_string_dtypes(t)]
+        ),
+        *(
+            pytest.param(
+                partial(t, [False, True, False], dtype=dt), [[3, 4]], id=f"{_n(t)}-{dt}"
+            )
+            for t in [pd.Index, pd.Series, pd.array]
+            for dt in ["bool", *([] if t is np.array else ["bool[pyarrow]"])]
+        ),
+    ],
+)
+def test_index_types(mk_idx: Callable[[], Index], expected: ad.AnnData) -> None:
+    index = mk_idx()
+    adata = ad.AnnData(X=np.array([[1, 2], [3, 4], [5, 6]]))
+    view = adata[index]
+    np.testing.assert_array_equal(view.X, np.array(expected))
+
+
 def test_ellipsis_index(
     ellipsis_index: tuple[EllipsisType | slice, ...] | EllipsisType,
     equivalent_ellipsis_index: tuple[slice, slice],
@@ -814,6 +882,22 @@ def test_index_3d_errors(index: tuple[int | EllipsisType, ...], expected_error:
         gen_adata((10, 10))[index]
 
 
+@pytest.mark.parametrize(
+    "index",
+    [
+        pytest.param(sparse.csr_matrix(np.random.random((1, 10))), id="sparse"),
+        pytest.param([1.2, 3.4], id="list"),
+        *(
+            pytest.param(np.array([1.2, 2.3], dtype=dtype), id=f"ndarray-{dtype}")
+            for dtype in [np.float32, np.float64]
+        ),
+    ],
+)
+def test_index_float_sequence_raises_error(index):
+    with pytest.raises(IndexError, match=r"has floating point values"):
+        gen_adata((10, 10))[index]
+
+
 # @pytest.mark.parametrize("dim", ["obs", "var"])
 # @pytest.mark.parametrize(
 #     ("idx", "pat"),
